'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var react = require('react');

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

/**
 * Create everything necessary to construct an accessible menu (or dropdown).
 * @param key - a key that uniquely identifies the menu on the current page
 * @param opts - options to further customize the menu's behavior
 */

function useMenu(key, opts) {
  var _opts$defaultIsOpen;

  // The main internal state of the menu
  var _useReducer = react.useReducer(reducer, opts != null && opts.defaultIsOpen ? _extends({}, INITIAL_STATE, {
    isOpen: (_opts$defaultIsOpen = opts == null ? void 0 : opts.defaultIsOpen) != null ? _opts$defaultIsOpen : false
  }) : INITIAL_STATE),
      state = _useReducer[0],
      dispatch = _useReducer[1]; // Create IDs for the menu and menu button


  var _useMemo = react.useMemo(function () {
    return ["use-menu-" + key, "use-menu-" + key + "-trigger"];
  }, [key]),
      menuId = _useMemo[0],
      buttonId = _useMemo[1]; // Initialize a helper to memoize event handlers of all menu items


  var _useMemoizedEventHand = useMemoizedEventHandlers(),
      nextMemoizedHandlers = _useMemoizedEventHand.nextMemoizedHandlers;

  return {
    isOpen: state.isOpen,
    buttonProps: {
      role: "button",
      id: buttonId,
      "aria-haspopup": "menu",
      "aria-expanded": state.isOpen ? "true" : undefined,
      "aria-controls": menuId,
      onClick: react.useCallback(function () {
        dispatch({
          type: "open"
        });
      }, [dispatch]),
      onKeyDown: react.useCallback(function (e) {
        switch (e.code) {
          case "Enter":
          case "Space":
          case "ArrowDown":
            dispatch({
              type: "open",
              focus: "first"
            });
            break;

          case "ArrowUp":
            dispatch({
              type: "open",
              focus: "last"
            });
            break;
        }
      }, [dispatch]),
      onBlur: react.useCallback(function (e) {
        // Close the menu, unless something inside the menu is focused
        var menu = document.getElementById(menuId);

        if (!menu || !e.relatedTarget || !menu.contains(e.relatedTarget)) {
          dispatch({
            type: "close"
          });
        }
      }, [dispatch, menuId])
    },
    menuProps: {
      role: "menu",
      id: menuId,
      "aria-labelledby": buttonId,
      ref: react.useCallback(function (menu) {
        if (!menu) {
          return;
        }

        switch (state.pendingFocus) {
          case "first":
            var first = menu.querySelector(SELECTOR_ITEMS);

            if (first) {
              first.focus();
            }

            break;

          case "last":
            var items = menu.querySelectorAll(SELECTOR_ITEMS);

            if (items.length > 0) {
              items[items.length - 1].focus();
            }

            break;
        }
      }, [state.pendingFocus]),
      onKeyDown: react.useCallback(function (e) {
        switch (e.code) {
          case "Escape":
            dispatch({
              type: "close"
            });
            break;

          case "ArrowDown":
            {
              var menu = e.currentTarget;
              var items = menu.querySelectorAll(SELECTOR_ITEMS);

              if (items.length === 0) {
                break;
              }

              for (var i = 1; i < items.length; ++i) {
                if (items[i - 1] === document.activeElement) {
                  items[i].focus();
                  return;
                }
              } // no next sibling found, wrap around


              items[0].focus();
            }
            break;

          case "ArrowUp":
            {
              var _menu = e.currentTarget;

              var _items = _menu.querySelectorAll(SELECTOR_ITEMS);

              if (_items.length === 0) {
                break;
              }

              for (var _i = _items.length - 2; _i >= 0; --_i) {
                if (_items[_i + 1] === document.activeElement) {
                  _items[_i].focus();

                  return;
                }
              } // no previous sibling found, wrap around


              _items[_items.length - 1].focus();
            }
            break;
        }
      }, [dispatch]),
      onBlur: react.useCallback(function (e) {
        var menu = e.currentTarget;

        if (!e.relatedTarget || !menu.contains(e.relatedTarget)) {
          dispatch({
            type: "close"
          });
        }
      }, [dispatch])
    },
    getItemProps: function getItemProps(callback, deps, opts) {
      var _nextMemoizedHandlers = nextMemoizedHandlers(function () {
        if (opts != null && opts.disabled) {
          return;
        }

        callback == null ? void 0 : callback();
        dispatch({
          type: "close"
        });
      }, function (e) {
        switch (e.code) {
          case "Enter":
          case "Space":
            if (opts != null && opts.disabled) {
              break;
            }

            callback == null ? void 0 : callback();
            dispatch({
              type: "close"
            });
            break;
        }
      }, deps),
          onClick = _nextMemoizedHandlers[0],
          onKeyDown = _nextMemoizedHandlers[1];

      return {
        role: "menuitem",
        tabIndex: -1,
        "aria-disabled": opts != null && opts.disabled ? true : undefined,
        onClick: onClick,
        onKeyDown: onKeyDown
      };
    },
    getItemCheckboxProps: function getItemCheckboxProps(opts) {
      var _nextMemoizedHandlers2 = nextMemoizedHandlers(function () {
        if (opts.disabled) {
          return;
        }

        opts.onToggle();
      }, function (e) {
        switch (e.code) {
          case "Enter":
          case "Space":
            if (opts.disabled) {
              break;
            }

            opts.onToggle();
            break;
        }
      }, [opts.onToggle]),
          onClick = _nextMemoizedHandlers2[0],
          onKeyDown = _nextMemoizedHandlers2[1];

      return {
        role: "menuitemcheckbox",
        tabIndex: -1,
        "aria-checked": opts.checked,
        "aria-disabled": opts.disabled ? true : undefined,
        onClick: onClick,
        onKeyDown: onKeyDown
      };
    }
  };
}
/**
 * A selector for all possible menu item types.
 */

var SELECTOR_ITEMS = "[role=menuitem],[role=menuitemcheckbox],[role=menuitemradio]";
/**
 * Determine whether two dependency lists are shallowly equal or not.
 */

function shallowEqual(lhs, rhs) {
  if (lhs.length !== rhs.length) {
    // If they have a different length, they are definitely not shallowly equal.
    return false;
  }

  for (var i = 0; i < lhs.length; ++i) {
    if (lhs[i] !== rhs[i]) {
      return false;
    }
  }

  return true;
}
/**
 * The initial internal state of the menu.
 */


var INITIAL_STATE = {
  isOpen: false,
  pendingFocus: null
};
/**
 * The reducer used to execute state changes.
 * @param state - the current state
 * @param action - the action applied on the current state
 */

function reducer(state, action) {
  switch (action.type) {
    case "open":
      // The open action acts like a toogle, if the menu is already open, it is closed
      if (!state.isOpen) {
        var _action$focus;

        return {
          isOpen: true,
          pendingFocus: (_action$focus = action.focus) != null ? _action$focus : null
        };
      } else {
        return {
          isOpen: false,
          pendingFocus: null
        };
      }

    case "close":
      return {
        isOpen: false,
        pendingFocus: null
      };
  }
}
/**
 * Memoize event handlers for an arbitrary length of menu items.
 */


function useMemoizedEventHandlers() {
  // The index of the next menu item. This is increased each time a set of event handlers is
  // memoized.
  var itemEventHandlerIndex = 0; // The array of memoized event handlers and their dependency list (which determines whether the
  // handlers have changed or not).

  var itemEventHandlers = react.useRef([]);
  return {
    /**
     * Get the next set of memoized event handlers (or update them if necessary).
     *
     * @param defaultOnClick - the default `onClick` handler, which is used if there is no memoized
     *    handler, or if the dependencies have changed
     * @param defaultOnKeyDown - the default `onKeyDown` handler, which is used if there is no
     *    memoized handler, or if the dependencies have changed
     * @param deps - the dependencies which are used to determine whether the handlers should be
     *    updated or not
     *
     * @returns a pair of event handlers, where the first one is the `onClick` and the second one
     *  is the `onKeyDown` handler
     */
    nextMemoizedHandlers: function nextMemoizedHandlers(defaultOnClick, defaultOnKeyDown, deps) {
      var _onClick, _onKeyDown;

      var memorized = itemEventHandlers.current[itemEventHandlerIndex];
      var onClick, onKeyDown; // Update the memoized handlers if the dependencies have changed

      if (memorized && deps) {
        var previousOnClick = memorized[0],
            previousOnKeyDown = memorized[1],
            previousDeps = memorized[2];

        if (shallowEqual(deps, previousDeps)) {
          onClick = previousOnClick;
          onKeyDown = previousOnKeyDown;
        }
      } // Always update the handlers if no dependencies are provided


      onClick = (_onClick = onClick) != null ? _onClick : defaultOnClick;
      onKeyDown = (_onKeyDown = onKeyDown) != null ? _onKeyDown : defaultOnKeyDown; // Memoized the updated event handlers

      itemEventHandlers.current[itemEventHandlerIndex] = [onClick, onKeyDown, deps ? deps.slice() : []]; // Increase the index so that the handlers for the next menu item are returned when
      // `nextMemoizedHandlers` is called again

      itemEventHandlerIndex++;
      return [onClick, onKeyDown];
    }
  };
}

/**
 * A hook to simplify keeping track of a checkbox form input state.
 *
 * @param initialState - the initial state of the checkbox
 */

function useMenuCheckboxState(initialState) {
  var _useState = react.useState(initialState),
      checked = _useState[0],
      setChecked = _useState[1];

  var onChange = react.useCallback(function (e) {
    return setChecked(e.target.checked);
  }, []);
  var onToggle = react.useCallback(function () {
    return setChecked(function (checked) {
      return !checked;
    });
  }, []);
  return {
    checked: checked,
    onToggle: onToggle,
    props: {
      checked: checked,
      onChange: onChange
    }
  };
}

exports.useMenu = useMenu;
exports.useMenuCheckboxState = useMenuCheckboxState;
//# sourceMappingURL=react-menu-hook.cjs.development.js.map
