import { KeyboardEventHandler, DependencyList, MouseEventHandler, FocusEventHandler, RefCallback } from "react";
/**
 * Options of the {@link useMenu} hook.
 */
export interface MenuOpts {
    /**
     * Whether the menu is already opened initially.
     */
    defaultIsOpen?: boolean;
}
/**
 * Create everything necessary to construct an accessible menu (or dropdown).
 * @param key - a key that uniquely identifies the menu on the current page
 * @param opts - options to further customize the menu's behavior
 */
export default function useMenu(key: string, opts?: MenuOpts): MenuState;
/**
 * All properties and event handlers necessary to construct an accessible menu.
 */
export interface MenuState {
    /**
     * Whether the menu is currently open or not.
     * @example
     * ```
     * {isOpen ? <div {...menuProps}>...</div>}
     * ```
     */
    isOpen: boolean;
    /**
     * The properties and event handlers for the button that opens and closes the menu.
     * @example
     * ```
     * <button {...buttonProps}>Open Menu</button>
     * ```
     */
    buttonProps: ButtonProps;
    /**
     * The properties and event handlers for the menu.
     * @example
     * ```
     * {isOpen ? <div {...menuProps}>...</div>}
     * ```
     */
    menuProps: MenuProps;
    /**
     * Get the properties and event handlers for the next menu item. Each call to this function refers
     * to one menu item.
     * @example
     * ```
     * <li {...getItemProps(() => {})}>Menu Action</li>
     * ```
     *
     * @param callback - the callback that is executed if the menu action is activated
     * @param deps - the optional dependencies that determine whether `callback` changed
     * @param opts - additional options for the menu item
     */
    getItemProps(callback?: () => void, deps?: DependencyList, opts?: ItemOptions): ItemProps;
    /**
     * Get the properties and event handlers for the next checkbox menu item. Each call to this
     * function refers to one checkbox menu item.
     * @example
     * ```
     * const checkbox = useMenuCheckboxState(false);
     * <li {...getItemCheckboxProps(const)}>
     *   <input type="checkbox" tabIndex={-1} {...checkbox.props} /> Checkbox
     * </li>
     * ```
     *
     * @param opts - options for the checkbox menu item
     */
    getItemCheckboxProps(opts: ItemCheckboxOptions): ItemCheckboxProps;
}
/**
 * The properties and event handlers for the button that opens and closes the menu.
 */
export interface ButtonProps {
    role: "button";
    id: string;
    "aria-haspopup": "menu";
    "aria-expanded"?: "true";
    "aria-controls": string;
    onClick: MouseEventHandler;
    onKeyDown: KeyboardEventHandler;
    onBlur: FocusEventHandler;
}
/**
 * The properties and event handlers for the menu.
 */
export interface MenuProps {
    role: "menu";
    id: string;
    "aria-labelledby": string;
    ref: RefCallback<HTMLElement | null>;
    onKeyDown: KeyboardEventHandler;
    onBlur: FocusEventHandler;
}
/**
 * The properties and event handlers for a menu item.
 */
export interface ItemProps {
    role: "menuitem";
    tabIndex: -1;
    "aria-disabled"?: true;
    onClick: MouseEventHandler;
    onKeyDown: KeyboardEventHandler;
}
/**
 * Additional options for a menu item.
 */
export interface ItemOptions {
    /**
     * Whether the menu item is disabled.
     */
    disabled?: boolean;
}
/**
 * The options for a checkbox menu item.
 */
export interface ItemCheckboxOptions {
    /**
     * Whether the checkbox is checked.
     */
    checked: boolean;
    /**
     * Whether the checkbox menu item is disabled.
     */
    disabled?: boolean;
    /**
     * A function to toggle the checkbox on and off.
     */
    onToggle(): void;
}
/**
 * The properties and event handlers for a checkbox menu item.
 */
export interface ItemCheckboxProps {
    role: "menuitemcheckbox";
    tabIndex: -1;
    "aria-checked": boolean;
    "aria-disabled"?: true;
    onClick: MouseEventHandler;
    onKeyDown: KeyboardEventHandler;
}
